/* drivers/barcelona/pwm/pwm.c
 *
 * Implementation of the pwm driver.
 *
 * Copyright (C) 2004,2005,2006 TomTom BV <http://www.tomtom.com/>
 * Authors: Dimitry Andric <dimitry.andric@tomtom.com>
 *          Jeroen Taverne <jeroen.taverne@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Includes */ 
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-timer.h>
#include <barcelona/timer.h>
#include <barcelona/Barc_pwm.h>
#include <barcelona/gopins.h>
#include <barcelona/gotype.h>
#include <barcelona/debug.h>
#include <barcelona/Barc_adc.h>

/* Defines */
#define PFX "pwm: "
#define PK_DBG PK_DBG_FUNC
#define SL_DBG PK_DBG_NONE

/* Enable to skip powering up the tps61042 chip. */
#undef PWM_SKIP_POWERUP

/* Internal stuff */
#define PWM_RANGE    (PWM_BACKLIGHT_MAX - PWM_BACKLIGHT_MIN)

static unsigned long pwm_tcnt;
static unsigned pwm_level = PWM_BACKLIGHT_MAX;

static unsigned short photosensor_data = 0;
static spinlock_t photosensor_data_lock = SPIN_LOCK_UNLOCKED;

static void photosensor_adc_poll(short buf[ADC_CHANNELS])
{
	unsigned long flags;

	/* Store photosensor data */
	spin_lock_irqsave(&photosensor_data_lock, flags);

	photosensor_data = buf[ADC_LX_OUT];

	spin_unlock_irqrestore(&photosensor_data_lock, flags);
}

static inline unsigned long divrnd(unsigned long nom, unsigned long den)
{
	return (nom + den / 2) / den;
}

static int pwm_hw_setlevel(unsigned level)
{
	unsigned long tcmp;
	static unsigned old_level = 0;

	IO_Activate(BACKLIGHT_EN);

	if (level < PWM_BACKLIGHT_MIN || level > PWM_BACKLIGHT_MAX) {
		PK_WARN("Invalid backlight level %u\n", level);
		return -EINVAL;
	}

	if (level == PWM_BACKLIGHT_MIN) {
		IO_Deactivate(BACKLIGHT_PWM);
		SL_DBG("level = %u, gpio pin off\n", level);
	} else if (level == PWM_BACKLIGHT_MAX) {
		IO_Activate(BACKLIGHT_PWM);
		SL_DBG("level = %u, gpio pin on\n", level);
	} else {
		/* Calculate PWM compare count */
		tcmp = divrnd((level - PWM_BACKLIGHT_MIN) * pwm_tcnt, PWM_RANGE);
		SL_DBG("level = %u, tcnt = %u, tcmp = %u\n", level, pwm_tcnt, tcmp);
		__raw_writel(tcmp, S3C2410_TCMPB(0));
#ifndef PWM_SKIP_POWERUP
		/* If the backlight was completely off before, we need to turn the GPB0
		 * pin on, and wait at least 500 us before switching to the PWM signal.
		 * See the TI TPS61042 specification for more information. */
		if (old_level == 0) {
			SL_DBG("powering up tps61042 chip...\n");
			IO_Activate(BACKLIGHT_PWM);
			udelay(500);
		}
#endif /* PWM_SKIP_POWERUP */
		IO_SetFunction(BACKLIGHT_PWM);
	}

	old_level = level;

	return 0;
}

static inline int pwm_setlevel(unsigned level)
{
	int ret;

	ret = pwm_hw_setlevel(level);
	if (ret < 0)
		return ret;

	pwm_level = level;
	return 0;
}

static int pwm_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case IOW_BACKLIGHT_OFF:
		return pwm_setlevel(PWM_BACKLIGHT_MIN);
	case IOW_BACKLIGHT_ON:
		return pwm_setlevel(PWM_BACKLIGHT_MAX);
	case IOW_BACKLIGHT_UPDATE:
		return pwm_setlevel(arg);
	case IOR_BACKLIGHT_CURRENT:
		return pwm_level;
	case IOR_LIGHTSENSOR_CURRENT:
		return photosensor_data;
	default:
		PK_WARN("Invalid ioctl command %u\n", cmd);
		return -EINVAL;
	}
}

static int pwm_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static int pwm_release(struct inode *inode, struct file *file)
{
	return 0;
}

/* Kernel interface */
static struct file_operations pwm_fops = {
	.owner		= THIS_MODULE,
	.ioctl		= pwm_ioctl,
	.open		= pwm_open,
	.release	= pwm_release,
};

static int pwm_hw_init(struct device *dev)
{
	struct clk *clock;
	unsigned long rate;
	unsigned long tcnt_mod;
	unsigned long flags;
	unsigned long tcon;
	unsigned long pwm_divider;
	unsigned long freq;

	clock = clk_get(dev, "pclk");
	if (IS_ERR(clock)) {
		PK_ERR("Failed to get PCLK\n");
		return -ENOENT;
	}
	rate = clk_get_rate(clock);
	PK_DBG("PCLK rate: %lu\n", rate);
	clk_put(clock);

	freq = IO_GetBacklightFreq();
	if (freq == 0) freq = 1000;
	pwm_divider = ((SYS_TIMER01_PRESCALER + 1) * SYS_TIMER0_DIVIDER * freq);

	pwm_tcnt = rate / pwm_divider - 1;
	tcnt_mod = rate % pwm_divider;
	PK_DBG("pwm_tcnt = %#lx, tcnt_mod = %#lx\n", pwm_tcnt, tcnt_mod);
	__raw_writel(pwm_tcnt, S3C2410_TCNTB(0));

	PK_DBG("Resetting backlight level\n");
//	pwm_hw_setlevel(pwm_level);

	PK_DBG("Starting timer 0\n");
	local_irq_save(flags);
	tcon = __raw_readl(S3C2410_TCON);
	tcon &= ~(0xf << 0);
	tcon |= S3C2410_TCON_T0RELOAD;
	tcon |= S3C2410_TCON_T0MANUALUPD;
	__raw_writel(tcon, S3C2410_TCON); /* Stop and manual load */

	tcon &= ~(0xf << 0);
	tcon |= S3C2410_TCON_T0RELOAD;
	tcon |= S3C2410_TCON_T0START;
	__raw_writel(tcon, S3C2410_TCON); /* Start */
	local_irq_restore(flags);
	
	PK_DBG("Done\n");
	return 0;
}

void pwm_restore_level(void)
{
	pwm_hw_setlevel(pwm_level);
}
EXPORT_SYMBOL(pwm_restore_level);

static void pwm_hw_exit(void)
{
	unsigned long flags;
	unsigned long tcon;

	PK_DBG("Turning off backlight\n");
	pwm_hw_setlevel(PWM_BACKLIGHT_MIN);

	PK_DBG("Stopping timer 0\n");
	local_irq_save(flags);
	tcon = __raw_readl(S3C2410_TCON);
	tcon &= ~(0xf << 0);
	tcon |= S3C2410_TCON_T0RELOAD;
	tcon |= S3C2410_TCON_T0MANUALUPD;
	__raw_writel(tcon, S3C2410_TCON); /* Stop and manual load */
	local_irq_restore(flags);
}

static int pwm_probe(struct device *dev)
{
	int ret;

	if (IO_HaveLightSensor())
	{
		PK_DBG("Registering ADC pollfunc\n");
		ret = adc_register_poll(photosensor_adc_poll);
		if (ret < 0) {
			PK_ERR("Unable to register ADC pollfunc (%d)\n", ret);
			return ret;
		}
	}

	ret = pwm_hw_init(dev);
	if (ret < 0) {
		PK_ERR("Failed to initialize hardware (%d)\n", ret);
		return ret;
	}

	PK_DBG("Registering chardev\n");
	ret = register_chrdev(PWM_MAJOR, PWM_DEVNAME, &pwm_fops);
	if (ret < 0) {
		PK_ERR("Unable to register chardev on major=%d (%d)\n", PWM_MAJOR, ret);
		return ret;
	}

	PK_DBG("Done\n");
	return 0;
}

static int pwm_remove(struct device *dev)
{
	pwm_hw_exit();

	if (IO_HaveLightSensor())
	{
		PK_DBG("Unregistering ADC poll\n");
		adc_unregister_poll(photosensor_adc_poll);
	}

	PK_DBG("Unregistering chardev\n");
	unregister_chrdev(PWM_MAJOR, PWM_DEVNAME);

	PK_DBG("Done\n");
	return 0;
}

static void pwm_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	pwm_hw_exit();
}

#ifdef CONFIG_PM

static int pwm_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (level == SUSPEND_POWER_DOWN) {
		pwm_shutdown(dev);
	}
	return 0;
}

static int pwm_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	if (level == RESUME_POWER_ON) {
		int ret = pwm_hw_init(dev);
		if (ret < 0) {
			PK_ERR("Failed to initialize hardware (%d)\n", ret);
			return ret;
		}
	}
	return 0;
}

#else /* CONFIG_PM */
#define pwm_suspend NULL
#define pwm_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver pwm_driver = {
	.name		= "tomtomgo-pwm",
	.bus		= &platform_bus_type,
	.probe		= pwm_probe,
	.remove		= pwm_remove,
	.shutdown	= pwm_shutdown,
	.suspend	= pwm_suspend,
	.resume		= pwm_resume,
};

static int __init pwm_mod_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO Backlight/Lightsensor Driver, (C) 2004,2005,2006 TomTom BV\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&pwm_driver);
	if (ret) {
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit pwm_mod_exit(void)
{
	PK_DBG("Unregistering driver\n");
	driver_unregister(&pwm_driver);
	PK_DBG("Done\n");
}

module_init(pwm_mod_init);
module_exit(pwm_mod_exit);

MODULE_AUTHOR("Dimitry Andric <dimitry.andric@tomtom.com> Jeroen Taverne <jeroen.taverne@tomtom.com>");
MODULE_DESCRIPTION("TomTom GO PWM Driver");
MODULE_LICENSE("GPL");

/* EOF */
